// -------------------------------------------------------
//
//  Copyright 2012, Richard L. Marker, All rights reserved
//
// -------------------------------------------------------

unit DonutAngleCalc;
interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Dialogs;

  {-----------------------------------------}
  {-----------------------------------------}
  {                                         }
  { The following items are maintained in   }
  { a list.  They are directly written to   }
  { the output set to be viewed by the      }
  { person running the program.             }
  {                                         }
  { The mapping to output is as follows:    }
  {   BaseLinks      ==> Normal             } 
  {   TwistedLinks   ==> Twisted            }
  {   BestLups       ==> Circle             }
  {   BestNum        ==> Loops              }
  {   BestCollAngle  ==> Coll Angle * 10^8  }
  {   BestFit        ==> Stability Measure  }
  {                                         }
  {-----------------------------------------}
  {-----------------------------------------}
  
type
  TBestItem = Class(TCollectionItem)
  private
    eBaseLinks,
    eTwistedLinks,
    eBestLups,
    eBestNum,
    eBestCollAngle,
    eBestFit        : extended;
  public
 //   constructor Create(Collection: TCollection); override;
    property BaseLinks:extended read eBaseLinks write eBaseLinks;
    property TwistedLinks:extended read eTwistedLinks write eTwistedLinks;
    property BestLups:extended read eBestLups write eBestLups;
    property BestNum:extended read eBestNum write eBestNum;
    property BestCollAngle:extended read eBestCollAngle write eBestCollAngle;
    property BestFit:extended read eBestFit write eBestFit;
  end; //TBestItem

  TBestList = Class(TCollection)
  private
    function GetItem(Index:Integer):TBestItem;
    function Add(BLinks,TLinks,BLups,BNum,CollAngle,BFit : extended): TBestItem;
    function Insert(Index:integer;BLinks,TLinks,
                    BLups,BNum,CollAngle,BFit : extended): TBestItem;
  public
    procedure EmptyList();
    property Items[Index : Integer]:TBestItem read GetItem;
  end;//TBestList

  TDonutCalc = Class(TObject)
  public
    BestList     : TBestList;
  private
    iRevTestCnt : integer;
    eThresholdRevMatch : extended;
    eThresholdBestFit : extended;
    eThresholdCollAngle : extended;
    iBestCnt : integer;     //max number of best hits to show
    iStartLoops,            //first donut chain untwisted length
    iFinishLoops : integer; //last donut chain untwisted length
    iMeasureAngle,          //iMeasure... number of times the factor
    iMeasureLinks,          //  occurs in the Best Fit Measure calculation
    iMeasureTime,
    iMeasureCycles: integer;

    procedure SetRevTestCnt(TC:integer);
    procedure SetThresholdRevMatch(RM:extended);
    procedure SetThresholdBestFit(BF:extended);
    procedure SetThresholdCollAngle(CA:extended);
    procedure SetBestCnt(BC:integer);
    procedure SetStartLoops(SL:integer);
    procedure SetFinishLoops(FL:integer);
    procedure SetMeasureAngle(MA:integer);
    procedure SetMeasureLinks(ML:integer);
    procedure SetMeasureTime(MT:integer);
    procedure SetMeasureCycles(MC:integer);
    function  CalcBestFit (NumLinks, CollisionAngle,
                           ElapsedTime,Cycles:extended):extended;

  public
    constructor CreateNew();
    procedure DoDonutCalcs();
    property RevTestCnt:integer read iRevTestCnt write SetRevTestCnt;
    property ThresholdRevMatch:extended read eThresholdRevMatch
                                        write SetThresholdRevMatch;
    property ThresholdBestFit:extended read eThresholdBestFit
                                       write SetThresholdBestFit;
    property ThresholdCollAngle:extended read eThresholdCollAngle
                                         write SetThresholdCollAngle;
    property BestCnt:integer read iBestCnt
                             write SetBestCnt;
    property StartLoops:integer read iStartLoops
                                write SetStartLoops;
    property FinishLoops:integer read iFinishLoops
                                 write SetFinishLoops;
    property MeasureAngle:integer  read iMeasureAngle
                                   write SetMeasureAngle;
    property MeasureLinks:integer  read iMeasureLinks
                                   write SetMeasureLinks;
    property MeasureTime:integer   read iMeasureTime
                                   write SetMeasureTime;
    property MeasureCycles:integer read iMeasureCycles
                                   write SetMeasureCycles;


  end; //TDonutCalc

var
 DonutCalc:TDonutCalc;

implementation

(*  tests integer ratios of donut loops to produce
     extra twist due to pi/2/n where n=min to max  *)

{$N+}

var
  SameLup      : Array[1..200] of extended;
  MaxSame      : integer;
{------------------------------------}
{ This list manages the solution set }
{ keeping the most stable solutions  }
{ that satisfy the inputted ranges   }
{ with stability criteria applied.   }
{------------------------------------}

function TBestList.GetItem(Index:Integer):TBestItem;
begin
  result := inherited Items[Index] as TBestItem;
end;//TBestList.GetItem

function TBestList.Add(BLinks,TLinks,
                       BLups,BNum,CollAngle,BFit : extended): TBestItem;
begin
  result := inherited Add as TBestItem;
  result.BaseLinks := BLinks;
  result.TwistedLinks := TLinks;
  result.BestLups := BLups;
  result.BestNum := BNum;
  result.BestCollAngle := CollAngle;
  result.BestFit := BFit;
end;//TBestList.Add

function TBestList.Insert(Index:integer; BLinks,TLinks,
                          BLups,BNum,CollAngle,BFit : extended): TBestItem;
begin
  result := inherited Insert(Index) as TBestItem;
  result.BaseLinks := BLinks;
  result.TwistedLinks := TLinks;
  result.BestLups := BLups;
  result.BestNum := BNum;
  result.BestCollAngle := CollAngle;
  result.BestFit := BFit;
end;//TBestList.Insert
{------------------------------}
procedure TBestList.EmptyList();
begin
   while Count > 0
     do Items[Count-1].Destroy;
end; {procedure TBestList.EmptyList}
{---------------------------------}
{ The following procedures simply }
{ set the inputted variables.     }
{---------------------------------}
procedure TDonutCalc.SetRevTestCnt(TC:integer);
begin
  iRevTestCnt := TC;
end; //Procedure TDonutCalc.SetRevTestCnt

procedure TDonutCalc.SetThresholdRevMatch(RM:extended);
begin
  eThresholdRevMatch := RM;
end; //Procedure TDonutCalc.SetThresholdRevMatch

procedure TDonutCalc.SetThresholdBestFit(BF:extended);
begin
  eThresholdBestFit := BF;
end; //procedure TDonutCalc.SetThresholdBestFit
procedure TDonutCalc.SetThresholdCollAngle(CA:extended);
begin
  eThresholdCollAngle := CA;
end; //procedure TDonutCalc.SetThresholdCollAngle

procedure TDonutCalc.SetBestCnt(BC:integer);
begin
  iBestCnt := BC;
end;//procedure TDonutCalc.SetBestCnt

procedure TDonutCalc.SetStartLoops(SL:integer);
begin
  iStartLoops := SL;
end;//procedure TDonutCalc.SetStartLoops

procedure TDonutCalc.SetFinishLoops(FL:integer);
begin
  iFinishLoops := FL;
end;//procedure TDonutCalc.SetFinishLoops

procedure TDonutCalc.SetMeasureAngle(MA:integer);
begin
  iMeasureAngle := MA;
end;//procedure TDonutCalc.SetMeasureAngle

procedure TDonutCalc.SetMeasureLinks(ML:integer);
begin
  iMeasureLinks := ML;
end;//procedure TDonutCalc.SetMeasureLinks

procedure TDonutCalc.SetMeasureTime(MT:integer);
begin
  iMeasureTime := MT;
end;//procedure TDonutCalc.SetMeasureTime

procedure TDonutCalc.SetMeasureCycles(MC:integer);
begin
  iMeasureCycles := MC;
end;//procedure TDonutCalc.SetMeasureCycles

{---------------------------------------------------------------}
{--- The CalcBestFit test calculates a number that will be   ---}
{--- tested to determine the best result.  The number of     ---}
{--- times each factor is used is input on the form.         ---}
{--- The resulting factors are normalized based on the       ---}
{--- electron and multiplied by 1000.  This is arbitrary.    ---}
{--- The purpose of normalization is to allow a constant     ---}
{--- maximum test quantity to be used even if the relative   ---}
{--- factor weightings are changed.                          ---}
{---------------------------------------------------------------}
function TDonutCalc.CalcBestFit (NumLinks, CollisionAngle,
                                 ElapsedTime, Cycles:extended):extended;
const
  eMeasureAngle  : extended = 1.0/0.008087279826;
  eMeasureLinks  : extended = 1.0/138.0;
  eMeasureTime   : extended = 1.0/74445.0;
  eMeasureCycles : extended = 1.0/4.0;
var
  measure : extended;
  i : integer;
begin
  measure := 1000.0;
  for i := 1 to MeasureAngle
    do measure := measure * CollisionAngle * eMeasureAngle;
  for i := 1 to MeasureLinks
    do measure := measure * NumLinks * eMeasureLinks;
  for i := 1 to MeasureTime
    do measure := measure * ElapsedTime * eMeasureTime;
  for i := 1 to MeasureCycles
    do measure := measure * Cycles * eMeasureCycles;
  result := measure;
end; //function TDonutCalc.CalcBestFit
{---------------------------------------------------------------}

Constructor TDonutCalc.CreateNew;
begin
  inherited Create;
end; //Constructor TDonutCalc.Create
{-------------------------------------}
{ Test the solution to see if it has  }
{ already been included.              }
{-------------------------------------}
function SameTest(Lups : extended) : Boolean;
 var
   E1 : extended;
   I1 : integer;
 begin
   I1 := 1;
   while (I1<201) and (I1<=MaxSame) do
     begin
       E1 := Lups/SameLup[I1];
       while (E1 > 0.0001) do E1 := E1 - 1.0;
       if abs(E1)< 0.00000000001
         then I1 := 202
         else inc(I1);
     end;  {while}
   if I1 = 202
    then SameTest := TRUE
    else
     begin
      Inc(MaxSame);
      if (MaxSame<201)then SameLup[MaxSame] := Lups;
      SameTest := FALSE;
     end;
 end; {function SameTest}
{-------------------------------------------}
{  *** This is the calculation itself. ***  }
{-------------------------------------------}
procedure TDonutCalc.DoDonutCalcs();

const
  LISTALL    = FALSE;
var
  AddALink   : boolean; {change to true to add link to chain}
  BigLupJump,
  Lups,
  NaMultNb,
  NB,
  NA,
  NUse,
  TRat,
  TanIncr,
  TanIncrSum,
  Tang,
  Angle1,
  Angle2,
  FitTest,
  CollAngle,
  LupsTestCnt  : Extended;
  RevErr       : Extended;
  NumLinks     : Extended;
  TestHit,
  Num          : integer;
  AlreadyInList,
  AddIt       : boolean;
  InsertIndex : integer;

begin
 for AddALink := False to True do
 for Num := StartLoops to FinishLoops do
 begin {main num loop}
  NumLinks := Num;
  NB := Num-1;
  NA := Num+1;
  {-------------------------------------------------------}
  {--- calc TRAT = tan(pi/4+pi/2/NB)                   ---}
  {--- TRat is the Tangent of the new angle with twist ---}
  {--- needed to align paths in the elongated angle    ---}
  {--- direction (i.e. shortest side twist).           ---}
  {--- Donuts have 4 orientations, one direction       ---}
  {--- is the preferred direction to twist(short side).---}
  {--- For adding a donut to chain use minus 0.5/na to ---}
  {--- account for less twist due to extra dn in chain.---}
  {--- Note that pi/4 (45 degrees) is normal collision ---}
  {--- angle (pi/2 orientation difference between     ---}
  {--- adjacent donuts divided by two.                ---}
  {-------------------------------------------------------}
  if AddALink
    then NUse := NA
    else NUse := NB;
  if AddALink
    then Angle1 := Pi * (0.25 - 0.5/NA)
    else Angle1 := Pi * (0.25 + 0.5/NB);
    {Angle1 is the desired new angle for exact parallel  }
    {alignment where the paths of adjacent donuts contact}
  TRat := Sin(Angle1)/Cos(Angle1);
  {-------------------------------------------------------}
  {--- do two tests (n-1)*(n+1) increments and n       ---}
  {-------------------------------------------------------}
    NaMultNb := NA*NB;{NA*NB;}
    TanIncr     := TRat / NaMultNb;
    {------ start just below the first possible integer soln ------}
    Lups        := 0.0;
    TanIncrSum  := 0.0;
    MaxSame     := 0;
    {--------------------------------------------}
    {--- main loops                           ---}
    {--- This is a brute force solution.      ---}
    {--------------------------------------------}
 //LupsTestCnt is the arbitrary upper limit
    LupsTestCnt := RevTestCnt*Num;
    while (Lups < LupsTestCnt) do
     begin
       TanIncrSum := TanIncrSum + TanIncr;
       Lups := Lups + 1.0;
       {--------------------------------------------------}
       {--- two error condition tests:                 ---}
       {---   RevErr for integer revolutions           ---}
       {---   CollAngle for collision angle            ---}
       {--- Do all loops for loop count save the       ---}
       {--- BestCnt best fits of collision angle that  ---}
       {--- satisfy threshold revolution matching.     ---}
       {--- Measure fit as the lowest collision        ---}
       {--- angle multiplied by revolution count.      ---}
       {--- Call this BestFit.                         ---}
       {--------------------------------------------------}
       RevErr := TanIncrSum+0.5-2.0*TanIncr;
       RevErr := Round(RevErr);
       RevErr := RevErr - TanIncrSum;
       {------ short cut to bypass unneeded calcs -----}
       if (RevErr > 2.0*TanIncr) then
         begin
           BigLupJump := Round(1.0 / TanIncr)- 4;
           If BigLupJump < 1 then BigLupJump := 0;
           Lups := Lups + BigLupJump;
           TanIncrSum := TanIncrSum + BigLupJump*TanIncr;
         end; //if (RevErr > 2.0*TanIncr)
       {------ apply threshold revolution matching ----}
       RevErr := abs(RevErr)*NaMultNb;  {need total mismatch on curr rev}
       if (RevErr < ThresholdRevMatch) and not SameTest(Lups) then
         begin
           Tang        := Round(TanIncrSum*NaMultNb)/Lups;
           Angle2      := ArcTan(Tang);  {actual angle achieved to the midline}
           CollAngle   := (Angle2 - Angle1)*1.0E+8; {collision angle to avg}
           FitTest     := CalcBestFit(NumLinks,CollAngle,Lups,Round(TanIncrSum));
           TestHit     := 0;
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++
// The remainder of the logic simply controls what is displayed.
// It does not affect the calculations
// if already in list, skip
// check to see if this should be added to the list for later display
// if LISTALL, make sure at least one is on the list regardless of
// whether or not it meets the thresholds; otherwise, allow a max of
// BestCnt.  Keep the best values lowest in index number.
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++
           AlreadyInList := False;
           while (TestHit < BestList.Count) and
                 not AlreadyInList do
             with BestList.Items[TestHit] do
             begin
               if (abs(CollAngle-BestCollAngle) < abs(0.0000001*CollAngle))
                 then AlreadyInList := True;
               inc(TestHit);
             end; //while

           if not AlreadyInList then
             begin
               // See if new hit qualifies for possibly adding
               // Negative angles are not allowed since it is
               // the angle that drives the alignment process
               if (((FitTest<ThresholdBestFit) and
                    (CollAngle>=0) and
                  (CollAngle<ThresholdCollAngle)) or
                  ((BestList.Count=0) and LISTALL))
                 then AddIt := True
                 else AddIt := False;
               // Find the Existing Item that is to be shifted up
               // or replaced
               InsertIndex := -1;
               TestHit := 0;
               while (TestHit < BestList.Count) and
                     (InsertIndex=-1) do
               begin
                 with BestList.Items[TestHit] do
                     if abs(FitTest)<abs(BestFit)
                       then InsertIndex:=TestHit;
                 inc(TestHit);
               end; //while (TestHit < BestList.Count)

               // do not add if not better and at max count
               if (InsertIndex=-1) and (BestCnt=BestList.Count)
                 then AddIt := False;

               // delete one if at max
               //add or insert the record
               if AddIt and (BestCnt=BestList.Count)
                then BestList.Delete(BestList.Count-1);

               if AddIt then
                 if (InsertIndex=-1)
                  then BestList.Add(Num,NUse,Lups,Round(TanIncrSum),
                         abs(CollAngle),CalcBestFit(NumLinks,
                                              CollAngle,Lups,Round(TanIncrSum)))
                  else BestList.Insert(InsertIndex,Num,Nuse,Lups,Round(TanIncrSum),
                       abs(CollAngle),CalcBestFit(NumLinks,
                                              CollAngle,Lups,Round(TanIncrSum)));

             end; //if not AlreadyInList
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++
         end; //if (RevErr < ThresholdRevMatch)
     end; {While Lups}
 end;  {main num loop}
end; {Procedure TDonutCalc.DoDonutCalcs}


end.

